#!/usr/bin/perl
#
# fraglets0.17/bin/cross.pl
#
# crossover operator (intended to replace cross.c)
#
# Copyright (C) 2005 <lidia.yamamoto@unibas.ch>
#
# This file is distributed under the terms of the GNU general Public
# License (GPL), see the file COPYING
#
# Usage:
#
# input (stdin) is a pair of genotypes
# output (stdout) is the new pair of genotypes
#
# cross over instructions:
#   cross.pl -<n> -f --maxs=<m> 
#
# <n> = 1:  one-point crossover
# <n> = 2 : two-point crossover
#
# -f  = cross only at fraglet border
#
# <m> = maximum segment size (in symbols; in fraglets if -f option)
#
# Note from biology: the output from a crossover is *not* an offspring,
# since crossover occurs during meiosis (i.e. crossover occurs between
# genes of the same individual).
#
# History:
# 20050201 rewritten from cross.c by lidia.yamamoto@unibas.ch
#

$usage="cross.pl -f --maxs=<m> <n>\n";

$fborder=$maxsize=$0;
$npoints=1;

while (@ARGV) {
    if($ARGV[0] eq "-f") {
	$fborder=1;
	shift;
    } elsif($ARGV[0] =~ /^--maxs=([0-9]+)$/) {
        $maxsize=$1;
        shift;
    } elsif($ARGV[0] eq "-1") {
        $npoints=1;
	shift;
    } elsif($ARGV[0] eq "-2") {
	$npoints=2;
	shift;
    } else {
        print STDERR "cross.pl: unknown argument $ARGV[0]\n$usage";
	exit -1;
    }
}

print "f=$fborder maxs=$maxsize npoints=$npoints\n";

sub parse_frag {
    local($node,$fcore,$mult) = @_;
    if ($mult > 0) {
	local(@symlist) = split(/:/, $fcore);
	local($nsyms) = $#symlist + 1;
	#print "node=$node fcore=<$fcore> mult=$mult nsyms=$nsyms\n";
	$frec[$ig][$fc]{node} = $node;
	@{ $frec[$ig][$fc]{symlist} } = @symlist;
	$frec[$ig][$fc]{nsyms} = $nsyms;
	$frec[$ig][$fc]{mult} = $mult;
    
	$fc[$ig] ++;
	$tnsyms[$ig] += $nsyms * $mult;
    }

    # should mult itself counts as one symbol?
    # $tnsyms += $nsyms * $mult + 1;
}

sub parse {
    while (<>) {
	if ( /^g[ \t]+\/(.*)\/[ \t]*$/ ) {
	    $gencore=$1;
	    @{$flist[$ig]} = split(/\//, $gencore);
	    $fc[$ig] = 0;
	    $tnsyms[$ig] = 0;
	    foreach $frag (@{$flist[$ig]}) {
		print "frag=$frag\n";
		if ($frag =~  /(.*)\[(.*)\]([0-9]+)/ ) {
		    parse_frag($1,$2,$3);
		}
	    }
	    return if (++$ig == 2);
	}
    }
}

sub print_genotype {
    local($ig) = @_;
    print "g /";
    for ($i=0; $i < $fc[$ig]; $i++) {
	print "$flist[$ig][$i]/";
    }
    print "\n";
}


sub min {
    local($x,$y)=@_;
    return ($x < $y ? $x : $y);
}

sub cross1p {
    $rndmax = min($tnsyms[0], $tnsyms[1]);
    $crosspoint = int (rand() * ($rndmax-1));
    print "crosspoint=$crosspoint\n";
}

sub cross2p {
    $rndmax = min($tnsyms[0], $tnsyms[1]);
    $crosspoint = int (rand() * ($rndmax-1));
    print "crosspoint=$crosspoint\n";
}

sub crossover {
    print "ig=$ig, fc[0]=$fc[0], fc[1]=$fc[1] tnsyms[0]=$tnsyms[0] tnsyms[1]=$tnsyms[1]\n";
    &cross1p;
}


$ig=0;
$rptot=0;

&parse();
if ($ig == 2 && $fc[0] > 0 && $fc[1] > 0) {
    &crossover();
    &print_genotype(0);
    &print_genotype(1);
} else {
    print STDERR "mutate.pl: wrong input format\n";
}

