/* fraglets0.31/src/frag.c

  fraglet interpreter

  Copyright (C) 2003-2007 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  fraglet interpreter implementation

  History:

  20070924 cft: added [printsym sym tail] --> [tail]
           (and sym is printed to stdout)
  20070924 lidia: added abs operator, renamed 'diff' to 'sub' (subtract)
           [abs tag n tail] --> [tag abs(n) <tail>]
  20070921 lidia: added stdout/stderr options to send (for printing)
           forbid node and segment names with reserved keywords
  20070920 lidia: fixed bug in pop2 (was using SYM_EOL, shouldn't!!)
  20070919 lidia: added "mod" (%) operator
  20070919 lidia: new broadcast instruction:
           n[broadcast ch msg] --> m[msg]
	   where m are the nodes attached to n via channel ch.
  20070805 lidia: support to node hierarchy in f_newnode, plus two
           new transformations:
           n1[inject n2 tail] --> n2[tail] if n2 is a child of n1
	   n2[expel tail] --> n1[tail] where n1 is the parent of n2
  20070720  thomas: added "fork" and "eq" instructions
  20070404  daniel: minor bugfix in splitat function
  20070320  lidia: added "splitat" transformation
  20070314  lidia: added "pop2" transformation:
            [pop2 h t a b c] --> [h a], [t b c]
            this is useful to separate input parameters for processing
	    example:
	    [match myf pop2 first rest], [myf x1 x2] -->
	    [first x1], [rest x2]
	    now we can easily grab x1 and x2 for separate processing;
	    in case of several parameters it suffices to apply pop2
	    several times:
	    [match rest pop2 r1 r2], [rest x2 x3 x4] -->
	    [r1 x2], [r2 x3 x4]
	    
  20070314  daniel.schreckling@informatik.uni-hamburg.de: added negative
            number support
  20070312  lidia: added "pow" transformation (x^y: x to the power y)
  20070303  lidia: added "mult" and "div" transformations
            (multiplication and division)
            increased symbol size to long
  20070121  christian.tschudin@unibas.ch: added 'x' command to exit
            without executing

  20070114  lidia.yamamoto@unibas.ch: bug fix: node_addsegment fixed so
            that it does not reinsert existing segments in node.

  20070112  lidia.yamamoto@unibas.ch:
    * added "length" transformation that returns the length of a fraglet:
      [length t1 a b c d] --> [t1 4 a b c d]

  20070111  lidia.yamamoto@unibas.ch:
    * added "sum" transformation [sum t1 2 3 tail] --> [t1 5 tail]
    * added "newname" transformation:
      [newname t1 prefix suffix tail] --> [t1 prefixsuffix tail]
    * added circular buffer as a palliative for name_lookup static buf
      problem for numbers
    * bug fix: waiting list was not being deleted by f_cleanup, so
      waiting fraglets would jump from one execution to the next.
    * bug fix in seg_parse: "a a", or "a[something]" would lead to
      Segmentation fault

  20070110  lidia.yamamoto@unibas.ch:
    * print production side of transformation when wait/delay expires:
      ... --> [fraglet that was waiting]
      this is important when calculating the concentration of fraglets
    * comments now accepted at the end of "a " and "f " lines;
    * removed the need of "e" at EOF (i.e. consider EOF='e');
    * added dynamic node creation:
      [newnode <seg> <node> tail] --> a <node> <seg>, <node>[tail]

  20070104  lidia.yamamoto@unibas.ch and christian.tschudin@unibas.ch:
	    added "delay" and "copy" commands
  20061214  christian.tschudin@unibas.ch: removed need for ":" field sep,
            added the "empty" transformation
  20061213  christian.tschudin@unibas.ch: added number support, and
            the "lt" transformation
  20050526  Jasenko Zivanov (Unibas): added socket support
  20050220  lidia.yamamoto@unibas.ch: reset random seed for repeatability
  20050217  christian.tschudin@unibas.ch (fraglets v0.18):
    * fixed bug that didn't correctly take multiplicity into account for
      the probability of choosing a fraglet for reaction
    * improved debug and execution traces
  20050206  by author and lidia.yamamoto@unibas.ch (fraglets v0.17):
    * added (or recovered?) support for 'nul' operator
    * support for markers /m markername/ in genotype (selfman'05 and wac'05)
    * improved trace printout
  20040704  updated by christian.tschudin@unibas.ch for wac'04 (fraglets v0.15)
  200307??  created by christian.tschudin@unibas.ch (ains'03)
*/

#include "sockin.h"
#include "frag.h"

struct frag *fraglist;
struct node *nodelist;
struct name *namelist;
int linecnt;

int debug;
//	swap comments to disable logging at all levels
#define DEBUG(l, stmt) if (l <= debug) do { stmt } while (0)
//#define DEBUG(l, stmt) 
int dump;
int quiet;

// (lidia:) if experiment should be repeatable from one run to the next
int repeatable;

// J. Zivanov: loss emulation
int packet_loss_probability = 0;

extern char *name_lookup(symbol s);

static void f_print(FILE *fout, struct frag *f, int eol);

// ------------------------------------------------------------

// get symbol at position n in fraglet f
#define f_getsymbol(f,n)   ((f)->buf->data[(f)->offs + (n)])

// set symbol at position n to s, in fraglet f
#define f_setsymbol(f,n,s) ((f)->buf->data[(f)->offs + (n)] = (s))

#define name_isreserved(s) ((s) < (SYM_EOL))

int istrans(symbol s)
{
	return s > 0 && s <= SYM_LAST_TRANS;
}

int isnumber(symbol s)
{
	return s >= SYM_NUMBEROFFS;
}

// symbol to number conversion
number numberval(symbol s)
{
	return isnumber(s) ? s - SYM_NUMBEROFFS - NEGNUMBEROFFS : 0;
}

// number to symbol conversion (watch out for overflow)
symbol int2symbol(number val) {
	symbol s = SYM_NUMBEROFFS + NEGNUMBEROFFS + (val);
	return isnumber(s) ? s : int2symbol(0);
}


// ------------------------------------------------------------

symbol name_add(char *cp)
{
	struct name *n = namelist;
	symbol i = 1;

	if ( isdigit(*cp) || (cp[0] == '-' && isdigit(cp[1]))) {
		number val;
#ifdef T_SYM_LONG
		val = atol(cp);
#else
		val = atoi(cp);
#endif
		return int2symbol(val);
	}

	while (n && strcmp(n->cp, cp)) {
		n = n->next;
		i++;
	}
	if (!n) {
		n = calloc(1, sizeof(*n));
		n->next = namelist;
		namelist = n;
		n->cp = malloc(strlen(cp) + 1);
		strcpy(n->cp, cp);
		n->sym = i;
	}
	
	return n->sym;
}


/* (lidia)
   warning!!! name_lookup returns a static buffer for numbers:

   originally only one buffer position was available, so this wouldn't
   work: printf("%s, %s",name_lookup(number1), name_lookup(number2));

   I modified it to use 10 buffer positions in a circular way, so that
   up to 10 calls are supported without overwriting the buffer (this
   is a poor man's garbage collection...)
*/

#define NUMBER_BUFSIZE 10

char *name_lookup(symbol s)
{
	struct name *n;
	static int bidx = 0;
	static char buf[NUMBER_BUFSIZE][32];

	if (isnumber(s)) {
	        bidx = (bidx+1)%NUMBER_BUFSIZE;
#ifdef T_SYM_LONG
		sprintf(buf[bidx], "%ld", numberval(s));
#else
		sprintf(buf[bidx], "%d", numberval(s));
#endif
		return buf[bidx];
	} else {
	        for (n = namelist; n; n = n->next)
		        if (n->sym == s)
			  return n->cp;
		return 0;
	}
}


/* set_trigger: allows application to set new name for trigger tag
 */

symbol set_trigger(char *cp)
{
	struct name *n;

	if ( isdigit(*cp) || (cp[0] == '-' && isdigit(cp[1]))) {
	  fprintf(stderr,
		  "set_trigger: numeric trigger not supported: %s\n", cp);
	  exit(-1);
	}

	for (n = namelist; n && strcmp(n->cp, cp); n = n->next);

	if (n) {
	  if (!strcmp(n->cp, cp)) return n->sym; // no change, ignore
	  fprintf(stderr,
		  "set_trigger: reserved name %s not allowed as trigger\n",
		  cp);
	  exit(-1);
	}
	for (n = namelist; n && n->sym != SYM_TRIGGER; n = n->next);
	if (!n) {
	  fprintf(stderr,
		  "set_trigger: FATAL: local trigger not found\n",
		  cp);
	  exit(-1);
	}

	free(n->cp);
	n->cp = malloc(strlen(cp) + 1);
	strcpy(n->cp, cp);
	return n->sym;
}


// ------------------------------------------------------------

struct node* node_lookup(symbol node, int create)
{
	struct node *n = nodelist;

	while (n && n->name != node)
		n = n->next;
	if (create) {
	  fprintf(stderr, "FATAL: not supported\n");
          exit(-1);
	}
	return n;
}

struct node* node_create(symbol node, symbol parent)
{
	struct node *n, *p;
	n = node_lookup(node, 0);
	if (n) return n;

	if (name_isreserved(node)) {
	    DEBUG(9,
		  printf("## node_create: reserved keyword \"%s\" not allowed as a node name\n", name_lookup(node)););
	    DEBUG(9,
		  printf("## node_create: eol=%d,n=%d\n",SYM_EOL,node););
	    return NULL;
	}
	if (node != parent) {
	  p = node_lookup(parent, 0);
	  if (p->nchildren >= MAXNODES) {
	    DEBUG(9,
            printf("## node_create \"%s\" parent=\"%s\": too many nodes\n",
		   name_lookup(node), name_lookup(parent)););
	    return NULL;
	  }
	}

	n = calloc(1, sizeof(*n));
	n->name = node;
	n->udp_wrapper = 0;
	n->location = Local;
	n->next = nodelist;
	nodelist = n;

	n->parent = parent;
	if (node != parent) {
	  p->child[p->nchildren++] = node;
	}

	DEBUG(9, printf("## node_create \"%s\" parent=\"%s\" OK\n",
			name_lookup(node), name_lookup(parent)););

	return n;
}


// adds node at root level
struct node* node_add(symbol node)
{
	struct node *n, *p;
	symbol parent;
	n = node_lookup(node, 0);
	if (n) return n;

	parent = name_add("");
	if (!node_create(parent, parent)) {
	  fprintf(stderr,"FATAL: error creating root node\n");
	  exit(-1);
	}
	n = node_create(node, parent);
	if (!n) {
	  fprintf(stderr, "FATAL: node_add %s: failed\n",
		  name_lookup(node));
	  exit(-1);
	}
	return n;
}


int node_addsegment(struct node *n, symbol seg)
{
	int i;

	if (!n) return 0;

	if (name_isreserved(seg)) {
	  DEBUG(9, printf("## addsegment node=%s seg=%s FAILED: reserved keyword not allowed\n", name_lookup(n->name), name_lookup(seg)););
	  return 0;
	}

	DEBUG(9, printf("## addsegment node=%s seg=%s\n",
			name_lookup(n->name), name_lookup(seg)););

	// (lidia): do not reinsert existing segment
	for (i = 0; i < MAXSEGMENTS && n->segments[i] &&
	       n->segments[i] != seg; i++);
	if (i < MAXSEGMENTS) {
		n->segments[i] = seg;
	}
	else { // complain ...
          DEBUG(9, printf("## no segment place for seg %s in node %s\n",
			  name_lookup(seg), name_lookup(n->name)););
	  return 0;
	}
	return 1;
}

// (lidia) segment lookup: true if segment found in node
//
int seg_lookup(struct node *n, symbol seg)
{
	int i;

	if (!n) return 0;
	for (i = 0; i < MAXSEGMENTS && n->segments[i]; i++)
		if (n->segments[i] == seg)
			return 1;
	return 0;

}


/* ------------------------------------------------------------

b_new

creates a new fraglet symbol buffer and sets its length in symbols;
the actual symbols have to be written afterwards, by using f_setsymbol.

"ressurected" and updated by lidia, Wed Mar 14, 2007

*/

struct fragbuf* b_new(int len) // len in symbols
{
        struct fragbuf *newbuf;

	newbuf = (struct fragbuf *)
	  malloc(sizeof(struct fragbuf) + sizeof(symbol) * len);

	if (newbuf) {
	  newbuf->refcnt = 1;
	  newbuf->len = len;
	}
	return newbuf;
}

void b_free(struct fragbuf *buf)
{
	free(buf);
}


/* ------------------------------------------------------------

f_new: creates a new fraglet given:

b    = fraglet buffer where the symbols are stored
offs = offs pointer where fraglet processing should resume
len  = length of fraglet in symbols
mult = multiplicity counter
n    = node where fraglet resides

*/

struct frag* f_new(struct fragbuf *b, int offs, int len, int mult, symbol n)
{
	struct frag *f = (struct frag*) calloc(1, sizeof(*f));
	if (!f)
		return 0;
	// should check offs and len...
	f->buf = b;
	f->next = 0;
	f->nd = n;
	f->offs = offs;
	f->len = len;
	f->mult = mult;
	b->refcnt++;
	return f;
}

symbol f_peek(struct frag *f)
{
	if (f->offs + 1 > f->buf->len)
		return 0;
	else {
		return f->buf->data[f->offs];
	}
}


void f_free(struct frag *f)
{
	f->buf->refcnt--;
	if (f->buf->refcnt <= 0)
		b_free(f->buf);
	free(f);
}


void f_freeLinkedList(struct t_queue *q)
{
	struct frag *f;

	while (q->first) {
		f = q->first;
		q->first = f->next;
		f_free(f);
	}
	q->last = 0;
}


int ismatch(symbol s)
{
	return s == SYM_MATCH || s == SYM_MATCHP;
}

void t_append(struct frag *f)
{
	struct node *n = node_lookup(f->nd, 0);
	symbol h = f->buf->data[f->offs];

	DEBUG(9, printf("# append transformation %s\n", name_lookup(h)););

	f->next = 0;
	if (n->tq.last)
		n->tq.last->next = f;
	else
		n->tq.first = f;
	n->tq.last = f;
	//	addhint(n, h);
}


void w_append(struct frag *f, int duration)
{
	struct node *n = node_lookup(f->nd, 0);

	DEBUG(9, printf("# append wait %s %d\n",
			name_lookup(f->buf->data[f->offs]), duration););

	f->next = 0;

	// pending: modify wait_duration to real time instead of virtual

	f->expires = n->opcnt + duration;
	if (n->wait.last) {
		struct frag *p;

		if (n->wait.first->expires > f->expires) {
			f->next = n->wait.first;
			n->wait.first = f;
		} else {
			for (p = n->wait.first; p != n->wait.last; p = p->next) {
				if (p->next->expires > f->expires) {
					f->next = p->next;
					p->next = f;
					break;
				}
			}
			if (p == n->wait.last) {
				n->wait.last->next = f;
				n->wait.last = f;
			}
		}
	} else {
		n->wait.first = f;
		n->wait.last = f;
	}
}


void m_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'match' for symbol %s\n", name_lookup(h)););

	// (lidia): (temporarily): also accept numbers as match tags
	if (isnumber(h)) h = SYM_NUMBER;
	
	if (h >= MAXSYMBOLS) {
	  fprintf(stderr, "m_add: Array out of bound exception: mtab[%d]\n",h);
	  exit(-1);
	  return;
	}
	DEBUG(9, printf("## maxsym=%d, sym_noff=%d, h=%d\n",MAXSYMBOLS,SYM_NUMBEROFFS,h););

	f->next = 0;
	if (n->mtab[h].last)
		n->mtab[h].last->next = f;
	else
		n->mtab[h].first = f;
	n->mtab[h].last = f;
}


void o_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'other' for symbol %s\n", name_lookup(h)););

	// (lidia): also accept numbers as front symbols
	if (isnumber(h)) h = SYM_NUMBER;
	if (h >= MAXSYMBOLS) {
		fprintf(stderr, "o_add: Array out of bound exception: otab[%d]\n",h);
		exit(-1);
		return;
	}
	DEBUG(9, printf("## maxsym=%d, sym_noff=%d, h=%d\n",MAXSYMBOLS,SYM_NUMBEROFFS,h););

	f->next = 0;

	if (n->otab[h].last)
		n->otab[h].last->next = f;
	else
		n->otab[h].first = f;
	n->otab[h].last = f;
}

// receive a fraglet?
struct frag* rx_fraglet(struct frag *f) // , int offs, int len)
{
	symbol head;
	struct node *n;

	if (!f) return 0;

	n = node_lookup(f->nd, 0);

	if (!n) {
	  fprintf(stderr, "error: no node for fraglet [%s ...]\n",
		  name_lookup(f_peek(f)));
	  exit(-1);
	}
	DEBUG(9, printf("## rx_fraglet\n"););

	if (f->len == 0) {
		f_free(f);
		return 0;
	}
	
	head = f_peek(f);
	if (istrans(head)) {
		t_append(f);
	} else {
		if (ismatch(head)) {
			DEBUG(9, printf("## match\n"););
			if (f->len == 1) {
				f_free(f);
				return 0;
			}
			head = f->buf->data[f->offs + 1];
			m_add(n, head, f);
		} else
		        o_add(n, head, f);
	}
	n->possibleHint |= 1;
	return f;
}


struct frag* f_parsefrag(int lno, char *line)
{
	char *cp, *a, *b, *nname;
	int m;
	struct fragbuf *buf;
	symbol node, parent;
	struct node* n;

	// format: "poolname [ sym1 : sym2 : ... ] mult"

	a = strchr(line, '[');
	b = strchr(line, ']');
	if (!a || !b) {
		// this also captures markers of the form:  /m markername/
		return 0;
	}
	*a = *b = 0;

	cp = strtok(line, " \n\t\v\r");

	nname = line; // what remains of line is now the node name
	node = name_add(nname);
	node_add(node);

	cp = strtok(b+1, " \n\t\v\r");
	if (!cp || *cp == '#') // default multiplicity is 1
		m = 1;
	else
		m = atoi(cp);

	if (m <= 0) {
		fprintf(stderr, "parse error line %d --- multiplicity must be > 0\n", lno);
		return 0;
	}

	buf = calloc(1, sizeof(*buf));
	for (a = strtok(a+1, " :\n\t\v\r"); a; a = strtok(NULL, " :\n\t\v\r")) {
		buf->len++;
		buf = realloc(buf, sizeof(*buf) + buf->len * sizeof(symbol));
		buf->data[buf->len - 1] = name_add(a);
	}

	return f_new(buf, 0, buf->len, m, node);
}


// parse genotype
struct frag* f_parsegen(int lno, char *line)
{
	char *t;
	struct frag *flist = 0, *f;

	if (*line != '/')
		return 0;

	for (t = strtok(line+1, "/\n\r"); t ; t = strtok(NULL, "/\n\r")) {
		f = f_parsefrag(lno, t);
		if (f) {
			f->next = flist;
			flist = f;
		}
	}
	return flist;
}


symbol seg_parse(int lno, char *line, symbol *node, char* ip_address, int* port)
{
	char *cp;
	symbol s;

	// J. Zivanov: new segment format (UDP support):
	// "A <nodename> <segname> [(<IPv4 address>|"local") <port>]"

	if (strchr(line, '[') || strchr(line, ']')) {

	  // (lidia): if [ or ] in node/seg descr then they cannot be
	  // used in fraglets, so reject this line (probably this is a
	  // fraglet and the user forgot to start the line with 'f')

	        return 0;
	}

	cp = strtok(line+1, " \t\n");
	if (!cp || *cp == '#') {
		fprintf(stderr, "syntax error on line %d: missing node name\n", lno);
		return 0;
	}

	*node = name_add(cp);

	cp = strtok(0, " \t\n");
	if (!cp || *cp == '#') {
		fprintf(stderr, "syntax error on line %d: missing segment name\n", lno);
		return 0;
	}
	s = name_add(cp);

	cp = strtok(0, " \t\n");
	if (cp && *cp != '#')
	{
	  // UDP segment
		if (strcmp(cp, "local"))
		{
			memccpy(ip_address, cp, '\n', 15);
			ip_address[15] = '\n';
		}
		else ip_address[0] = 'l';
		cp = strtok(0, " \t\n");
		*port = atoi(cp);
		cp = strtok(0, " \t\n");
	}
	else ip_address[0] = '\0';
	
	if (!*node || !s || cp && *cp != '#') {
		fprintf(stderr, "syntax error on line %d\n", lno);
		return 0;
	}
	return s;
}


int f_import(FILE *fin)
{
	static int iseof = 0;

	char line[10000], *cp;
	struct frag *f, *f1, *f2;
	symbol node, seg;

	// J. Zivanov: udp support
	char ip_address[32];
	int port;

	if (iseof) return -1;
	f1 = NULL;
	while (fgets(line, sizeof(line)-1, fin)) {
		linecnt++;
		//if (line[0] == '\n') continue;
		for (cp = line; isspace(*cp); cp++);
		switch(*cp) {
		//case '\n':
		case '#':
		case '\0':
			continue;
		case 'e':
			return 0;
		case 'f':
		case 'F':
			for (cp++; isspace(*cp); cp++);
			f = f_parsefrag(linecnt, cp);
			if (!f) {
			  // Lidia: 'line' is broken at this stage, cannot be printed
				fprintf(stderr, "Line %d: cannot parse fraglet\n", linecnt);
				exit(-1);
			} else f1 = f;
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			continue; //break;
		case 'g':
		case 'G':
			for (cp++; isspace(*cp); cp++);
			f = f_parsegen(linecnt, cp);
			if (!f) {
				fprintf(stderr, "Line %d: cannot parse genotype\n", linecnt);
				exit(-1);
			}
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			return 0;
		case 'a':
		case 'A':
		  // J. Zivanov: added udp support for segment
			seg = seg_parse(linecnt, line, &node, ip_address, &port);
			if (!seg) {
				fprintf(stderr, "Line %d: cannot parse segment %s\n", linecnt, line);
				exit(-1);
			}
			if (!node_addsegment(node_add(node), seg)) {
			        fprintf(stderr, "FATAL: could not add segment %s\n", name_lookup(seg));
				exit(-1);
			}
			if (ip_address[0] != '\0') 
			{	
				if (ip_address[0] == 'l') open_local_socket(node, port);
				else open_socket_to_remote_node(node, ip_address, port);
			}
			continue; // break;
		case 'm':
		case 'M':
			continue; // break;
		case 'x':
			return -1;
		default:
			fprintf(stderr, "Line %d: cannot parse %s\n", linecnt, line);
			exit(-1);
		}
	}
	iseof = 1;
	if (f1) return 0;
	else return -1;
}


void f_compact_list(struct frag *f, struct frag **last)
{
	struct frag *f2, *f3, **fpp;

	while (f) {
		int len = f->len;
		symbol *sp = f->buf->data + f->offs;

		for (fpp = &f->next, f2 = f->next; f2; ) {
			if (len != f2->len ||
			    memcmp(sp, f2->buf->data + f2->offs,
				   sizeof(symbol) * len)) {
				fpp = &f2->next;
				f2 = f2->next;
				continue;
			}
			// same content!
			f->mult += f2->mult;
			f3 = f2->next;
			f_free(f2);
			f2 = *fpp = f3;
		}
		if (!f->next)
			*last = f;
		f = f->next;
	}
}


static void f_symprint(FILE *fout, struct frag *f, int eol)
{
	int i;
	for (i = 0; i < f->len; i++) {
		fprintf(fout, i ? " %s" : "%s",
			name_lookup(f_getsymbol(f,i)));
	}
	if (eol) fprintf(fout, "\n");
}


static void f_print(FILE *fout, struct frag *f, int eol)
{
	fprintf(fout, "%s[", name_lookup(f->nd));
	f_symprint(fout, f, 0);
	fprintf(fout, eol ? "]\n" : "]");
}


static void f_export_list(FILE *fout, struct node *n, struct frag *f)
{
	for (; f; f = f->next) {
		fprintf(fout, "f ");
		f_print(fout, f, 0);
		fprintf(fout, "%d\n", f->mult);
	}
}


static void f_export_mtab(FILE *fout, struct node *n, struct m_rec *mtab)
{
	int i;

	for (i = 0; i < MAXSYMBOLS; i++, mtab++) {
		f_compact_list(mtab->first, &(mtab->last));
		f_export_list(fout, n, mtab->first);
	}
}


int f_export(FILE *fout)
{
	struct node *n;

	for (n = nodelist; n; n = n->next) {
		int i;
		char *nname = name_lookup(n->name);
		if (n->segments[0]) {
			for (i = 0; n->segments[i]; i++) {
				fprintf(fout, "a %s %s\n", nname,
				name_lookup(n->segments[i]));
			}
		}
	}

	for (n = nodelist; n; n = n->next) {
		f_compact_list(n->tq.first, &(n->tq.last));
		f_export_list(fout, n, n->tq.first);
		f_compact_list(n->wait.first, &(n->wait.last));
		f_export_list(fout, n, n->wait.first);
		f_export_mtab(fout, n, n->mtab);
		f_export_mtab(fout, n, (struct m_rec*) n->otab);

	}

	return 0;
}


/* f_shrink removes the given number of front symbols. If
   empty, 0 is returned. Otherwise a fraglet (copy) is returned
   with a multiplicity of 1. Note that this copy still might
   share the buffer with some other fraglet, so you should not
   modify the content.
*/
struct frag* f_shrink(struct frag *f, int cnt)
{
	DEBUG(9, printf("# f_shrink %d\n", f->mult););

	if (f->mult == 1) {
		if (f->len <= cnt) {
			f_free(f);
			return 0;
		}
		f->offs += cnt;
		f->len -= cnt;
	} else {
		f->mult--;
		rx_fraglet(f);
		if (f->len > cnt)
			f = f_new(f->buf, f->offs + cnt,
				  f->len - cnt, 1, f->nd);
	}

	return f;
}


/* (lidia): copy a fraglet buffer (nasty f_new() doesn't do it!!!)
  such that it can be safely updated (after f_shrink does its dirty
  job); ignores already consumed headers, so buffer might be smaller
  than the original one. when the fraglet is already standalone
  (multiplicity is one and no one else is pointing to its buffer
  (refcnt=1) then fraglet is not changed.
*/
struct frag *f_bufcpy(struct frag *f) {
        struct fragbuf *newbuf;

	if (f->mult == 1 && f->buf->refcnt == 1) return f;
	if (f->mult > 1) f = f_shrink(f, 0);

	newbuf = b_new(f->len);

	if (!newbuf) {
	  fprintf(stderr, "f_bufcpy: fatal error: out of memory\n");
	  exit(-1);
	}

	memcpy(newbuf->data, f->buf->data + f->offs, sizeof(symbol) * f->len);
	
	f->buf->refcnt--;
	if (f->buf->refcnt <= 0) b_free(f->buf);
	f->buf = newbuf;
	f->offs = 0;
	return f;
}


// ----------------------------------------------------------------------

struct frag* f_nop(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	// Lidia: added execution trace
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	f = f_shrink(f, 1);

	if (f) {
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else
		DEBUG(1, printf("[]\n"););

	return f;
}


// lidia: copy entire fraglet
struct frag* f_copy(struct frag *f)
{
	DEBUG(9, printf("# f_copy %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	if (f->len < 2)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 1;
	f = f_shrink(f, 1);
	if (f) {
		f->mult++;
		node_lookup(f->nd, 0)->symbalance -= f->len;
		DEBUG(1, f_print(stdout, f, 0););
		DEBUG(1, printf("%d\n", f->mult););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// thomas: copy fraglet and prepend different header symbols
// [fork a b tail] --> [a tail], [b tail]
struct frag* f_fork(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_fork %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance -= f->len - 2 - 2;
	
	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data+1, f->buf->data + f->offs + 3,
	       sizeof(symbol) * (f->len - 3));
	buf->data[0] = f->buf->data[f->offs + 1];
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 0););
	rx_fraglet(f2);

	f = f_shrink(f, 2);
	if (f) {
		node_lookup(f->nd, 0)->symbalance -= f->len;
		DEBUG(1, printf(", "););
		DEBUG(1, f_print(stdout, f, 0););
		DEBUG(1, printf("\n", f->mult););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_delay(struct frag *f)
{
	DEBUG(9, printf("# f_delay %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> ...\n"););
	      
	if (f->len < 3)
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 1]))
		goto drop;

	f = f_shrink(f, 0);
	if (f)
		w_append(f, numberval(f->buf->data[f->offs + 1]));

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;

}


struct frag* f_dup(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_dup %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	if (f->len < 3)
		goto drop;

// dup is symbol count neutral:
// 	node_lookup(f->nd, 0)->symbalance += 0;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->buf->data[f->offs] = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * f->len);
	buf->len = f->len;
	buf->refcnt = 0;
	memcpy(buf->data + 2, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);

	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_empty(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;
	symbol s;

	DEBUG(9, printf("# f_empty %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->len > 3)
		s = f->buf->data[f->offs + 2];
	else
		s = f->buf->data[f->offs + 1];

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
		f->buf->data[f->offs] = s;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data+1, f->buf->data + f->offs + 3,
	       sizeof(symbol) * (f->len - 3));
	buf->data[0] = s;
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_exch(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_exch %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 4)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 1;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		f->offs++;
		f->len--;
		s = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		f->buf->data[f->offs + 2] = s;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 1));
	buf->len = f->len - 1;
	buf->refcnt = 0;
	memcpy(buf->data + 3, f->buf->data + f->offs + 4,
	       sizeof(symbol) * (f->len - 4));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 3];
	buf->data[2] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len - 1, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// (lidia)
// 
// [length tag <tail>] --> [tag <taillength> <tail>]
//
struct frag* f_length(struct frag *f)
{
   symbol tag, tlen;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 2)
		goto drop;
	
	tag = f_getsymbol(f,1);

	tlen = int2symbol(f->len - 2);

        f = f_bufcpy(f);
	f_setsymbol(f,0,tag);
	f_setsymbol(f,1,tlen);
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// (lidia)
// 
// [abs tag n tail] --> [tag abs(n) <tail>]
//
struct frag* f_abs(struct frag *f)
{
   symbol tag, ns;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;
	
	tag = f_getsymbol(f,1);
	ns = f_getsymbol(f,2);
	if (!isnumber(ns)) goto drop;
	ns = int2symbol(abs(numberval(ns)));
	f = f_shrink(f,1);
	if (f) {
	        f = f_bufcpy(f);
		//f_setsymbol(f,0,tag);  //already there
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


/* ------------------------------------------------------------

f_op2: arity 2 operators: sum, diff, mult, div

[sum  tag n1 n2 tail] --> [tag (n1+n2) tail]
[mult tag n1 n2 tail] --> [tag (n1*n2) tail]
[diff tag n1 n2 tail] --> [tag abs(n1-n2) tail]
[div  tag n1 n2 tail] --> [tag (n1/n2) tail] # fraglet discarded if n2==0
[pow  tag n1 n2 tail] --> [tag (n1^n2) tail]

*/

struct frag* f_op2(struct frag *f)
{
	symbol op, tag, n1, n2, ns;
	number val1, val2;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;

	op  = f_getsymbol(f,0);
	tag = f_getsymbol(f,1);
	n1  = f_getsymbol(f,2);
	n2  = f_getsymbol(f,3);
	if (!isnumber(n1) || !isnumber(n2)) goto drop;

	switch (op) {
	case SYM_SUM:
	  ns = int2symbol(numberval(n1) + numberval(n2)); break;
	case SYM_SUB:
	  ns = int2symbol(numberval(n1) - numberval(n2)); break;
	case SYM_MULT:
	  //DEBUG(8, printf("# mult %lu %lu = %lu\n", numberval(n1), numberval(n2), numberval(n1)*numberval(n2)););

	  ns = int2symbol(numberval(n1) * numberval(n2)); break;
	case SYM_DIV:
	  val2 = numberval(n2);
	  if (val2 == 0) goto drop; // discard fraglet if division by zero
	  ns = int2symbol(numberval(n1) / val2); break;
	case SYM_MOD:
	  val2 = numberval(n2);
	  if (val2 == 0) goto drop; // discard fraglet if division by zero
	  ns = int2symbol(numberval(n1) % val2); break;
	case SYM_POW:
	  val1 = numberval(n1);
	  val2 = numberval(n2);
	  if (val2 < 0 || (val1 == 0 && val2 == 0)) goto drop;
	  ns = int2symbol((number) pow((double) val1, (double) val2));
	  break;
	default:
	  fprintf(stderr, "unknown operator '%s'\n", name_lookup(op));
	  exit(-1);
	}

	f = f_shrink(f,2);
	if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,tag);
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// (lidia)
// [newname tag s1 s2 tail] --> [tag s1s2 tail]
//
struct frag* f_newname(struct frag *f)
{
	symbol tag, s1, s2, ns;
	char *str1, *str2, *newstr;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;
	
	tag = f_getsymbol(f,1);
	s1  = f_getsymbol(f,2);
	s2  = f_getsymbol(f,3);
	str1 = name_lookup(s1);
	str2 = name_lookup(s2);
	newstr = (char *) malloc(strlen(str1) + strlen(str2) + 1);
	strcpy(newstr,str1);
	strcat(newstr,str2);
	ns = name_add(newstr);
	free(newstr);

	f = f_shrink(f,2);
	if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,tag);
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_newnode(struct frag *f)
{
	symbol seg, newnode, parent;
	struct node *n, *p;

	parent = f->nd;
	p = node_lookup(parent, 0);
	DEBUG(1, printf("%05d  ", p->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;
	
	seg = f_getsymbol(f,1);
	newnode = f_getsymbol(f,2);

	n = node_lookup(newnode, 0);
	if (!n) {
	  DEBUG(9, printf("## newnode create %s parent=%s\n",
			  name_lookup(newnode), name_lookup(parent)););

	  n = node_create(newnode, parent);
	  if (!n) goto drop;
	  DEBUG(1, printf("## newnode create OK\n"););
	}

	DEBUG(9, printf("## newnode %s seg=%s parent=%s (%d)\n",
			name_lookup(newnode), name_lookup(seg),
			name_lookup(parent),
			f->mult););

	// invalid segment will not be added but node will be created anyway
	if (!node_addsegment(n, seg)) {
	  DEBUG(9, printf("## seg=%s not created, but proceeding\n", name_lookup(seg)););
	}

	f = f_shrink(f,3);
	if (f) {
		f->nd = newnode;
		node_lookup(f->nd, 0)->symbalance -= f->len;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_nul(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


/* 
    [pop h a b c] --> [h b c]

    (ly): TO DO: rename this "tail"

*/

struct frag* f_pop(struct frag *f)
{
	DEBUG(9, printf("# f_pop %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
		f->buf->data[f->offs] = f->buf->data[f->offs - 1];
	} else {
		struct fragbuf *b2;
		struct frag *f2;

		b2 = malloc(sizeof(*b2) + sizeof(symbol) * (f->len - 2));
		b2->len = f->len - 2;
		b2->refcnt = 0;
		memcpy(b2->data + 1, f->buf->data + f->offs + 3,
		       sizeof(symbol) * (f->len - 3));
		b2->data[0] = f->buf->data[f->offs + 1];
		f2 = f_new(b2, 0, f->len - 2, 1, f->nd);
		if (--(f->mult) <= 0)
			f_free(f);
		else {
			t_append(f);
//			addhint(node_lookup(f->nd, 0), SYM_POP);
		}
		f = f2;
	}
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

/* 
    [pop2 h t a b c] --> [h a], [t b c]
*/


struct frag* f_pop2(struct frag *f)
{
	symbol h, t, hs;
	struct frag *fh;
        struct fragbuf *bh;

	DEBUG(9, printf("# f_pop2 len=%d, mult=%d\n", f->len, f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 3)
		goto drop;

	h  = f_getsymbol(f,1);
	t  = f_getsymbol(f,2);
	if (f->len < 4)
	  hs = 0; // no symbol
	else
	  hs = f_getsymbol(f,3);

	// creates [h a] fraglet
	bh = b_new(hs == 0 ? 1 : 2);
	fh = f_new(bh, 0, bh->len, 1, f->nd);
	if (fh) {
	  f_setsymbol(fh,0,h);
	  if (hs) f_setsymbol(fh,1,hs);
	  DEBUG(1, f_print(stdout, fh, 0); printf(", "););
	  rx_fraglet(fh);
	}

	if (f->len < 4) {
	  // [pop2 h t] --> [h], [t]
	  f = f_shrink(f, 2);
	}
	else {
	  // modify current fraglet to contain [t b c ...]
	  f = f_shrink(f, 3);
	  if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,t);
	  }
	}
	if (f) {
	  DEBUG(1, f_print(stdout, f, 1););
	  rx_fraglet(f);
	} else {
	  DEBUG(1, printf("[]\n"););
	}
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


/*
   [printsym sym tail] --> [tail]
      print sym to stdout

*/

struct frag* f_printsym(struct frag *f)
{
	struct node *n = node_lookup(f->nd, 0);
	char *nn;
	symbol sym;

	DEBUG(9, printf("# f_printsym %d\n", f->mult););
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 3)
		goto drop;

	sym = f->buf->data[f->offs+1];
	n->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
	} else {
		struct fragbuf *b2;
		struct frag *f2;

		b2 = malloc(sizeof(*b2) + sizeof(symbol) * (f->len - 2));
		b2->len = f->len - 2;
		b2->refcnt = 0;
		memcpy(b2->data, f->buf->data + f->offs + 2,
		       sizeof(symbol) * (f->len - 2));
		f2 = f_new(b2, 0, f->len - 2, 1, f->nd);
		if (--(f->mult) <= 0)
			f_free(f);
		else {
			t_append(f);
		}
		f = f2;
	}
	DEBUG(1, f_print(stdout, f, 1););

	nn = name_lookup(f->nd);
	if (!*nn)
		printf("# default node: %s\n", name_lookup(sym));
	else
		printf("# node %s: %s\n", nn, name_lookup(sym));

	rx_fraglet(f);
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_split(struct frag *f)
{
	symbol *sp;
	int i;
	struct frag *f2;

	DEBUG(9, printf("# f_split %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	for (i = 1, sp = f->buf->data + f->offs + 1; i < f->len; i++, sp++) {
		if (*sp == SYM_STAR)
			break;
	}
	if (i == 1) { // empty first half
		node_lookup(f->nd, 0)->symbalance += 2;
		f = f_shrink(f, 2);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
			
		rx_fraglet(f);
		return f;
	}
	if (i >= f->len) { // no star (empty second half)
		node_lookup(f->nd, 0)->symbalance += 1;
		f = f_shrink(f, 1);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
		rx_fraglet(f);
		return f;
	}
	node_lookup(f->nd, 0)->symbalance += 2;
	if (i != (f->len - 1)) { // accept the second half, if not empty
		f2 = f_new(f->buf, f->offs + i + 1, f->len - i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 0); printf(", "););
		rx_fraglet(f2);
	}
	// handle the first half
	if (f->mult == 1) {
		f->offs += 1;
		f->len = i - 1;
		DEBUG(1, f_print(stdout, f, 1););
	} else {
		f->mult--;
		f2 = f_new(f->buf, f->offs + 1, i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 1););
		rx_fraglet(f2);
	}
	rx_fraglet(f);
	return 0;
}


/*

splitat: splits at the first occurrence of symbol h, which is the
symbol following the splitat tag. example:

[splitat h a b c h x y z] --> [a b c], [x y z]

warning!! splitat is *not* a O(1) operation!

 */

struct frag* f_splitat(struct frag *f)
{
	symbol hs;
	int i;
	struct frag *f2;

	DEBUG(9, printf("# f_splitat %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 2)
	  goto drop;

	hs = f_getsymbol(f,1);

	for (i = 2; i < f->len; i++) {
	  if (f_getsymbol(f,i) == hs)
	    break;
	}
	DEBUG(8, printf("## f_splitat i=%d len=%d\n",i,f->len););
	if (i == 2) { // empty first half
		f = f_shrink(f, 3);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
			
		rx_fraglet(f);
		return f;
	}
	if (i >= f->len) { // symbol not found (empty second half)
		f = f_shrink(f, 2);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
		rx_fraglet(f);
		return f;
	}
	if (i != (f->len - 1)) { // accept the second half, if not empty
		f2 = f_new(f->buf, f->offs + i + 1, f->len - i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 0); printf(", "););
		rx_fraglet(f2);
	}
	// handle the first half
	if (f->mult == 1) {
		f->offs += 2;
		f->len = i - 2;
		DEBUG(1, f_print(stdout, f, 1); );
	} else {
		f->mult--;
		f2 = f_new(f->buf, f->offs + 2, i - 2, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 1););
		rx_fraglet(f2);
	}
	rx_fraglet(f);
	return 0;
drop:
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// broadcast to channel
//
// n[broadcast ch msg] --> m[msg]
//
// where m are the nodes attached to n via channel ch.

struct frag* f_bcast(struct frag *f)
{
	symbol seg;
	struct node *nsrc, *ndst;
	struct frag *f2;
        struct fragbuf *bn;
	int prev;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;
	
	seg = f_getsymbol(f,1);
	nsrc = node_lookup(f->nd, 0);

	if (!seg_lookup(nsrc,seg)) goto drop;

	f = f_shrink(f,2);
	if (!f) {
	  DEBUG(1, printf("[]\n"););
	  return 0;
	}

	for (prev=0, ndst = nodelist; ndst ; ndst = ndst->next) {
	  if (ndst != nsrc && seg_lookup(ndst, seg)) {
	    DEBUG(9, printf("## broadcast from %s to %s\n",
		     name_lookup(nsrc->name), name_lookup(ndst->name)););
	    if (ndst->location == Local) {
	      bn = b_new(f->len);
	      memcpy(bn->data, f->buf->data + f->offs,
		     sizeof(symbol) * f->len);
	      f2 = f_new(bn, 0, bn->len, 1, ndst->name);

	      if (prev) {
		DEBUG(1, printf(", "););
	      }
	      DEBUG(1, f_print(stdout, f2, 0););
	      rx_fraglet(f2);
	    } else {
	      f->nd = ndst->name;
	      DEBUG(1, f_print(stdout, f, 0); printf(" (unreliably)\n"););
	      send_fraglet_through_socket(f, nsrc, ndst);
	    }
	    prev=1;
	  }
	}
	if (prev) {
	  DEBUG(1, printf("\n"););
	}

	// silently discard used fraglet at the end:
	goto finish;
drop:
	DEBUG(1, printf("[]\n"););
finish:
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;

}


// J. Zivanov: changed this to send to socket when specified in topology
// line of destination; must also check whether topology line of
// sender has a compatible port & addr combination, and whether they
// share the same "segment" ; actually we can specify that all nodes
// able to communicate via udp share the same segment - we could even
// call this segment just "udp".

struct frag* f_send(struct frag *f)
{
	symbol chan;
	FILE *cf;
	struct node *dest, *n;
	struct frag *f2;
	int i;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;
	
	chan = f_getsymbol(f,1);
	if (chan == SYM_STDOUT || chan == SYM_STDERR) {
	  cf = ((chan == SYM_STDOUT) ? stdout : stderr);
	  f = f_shrink(f,2);
	  f->nd = SYM_STDOUT;
	  DEBUG(1, f_print(stdout, f, 1););
	  f_symprint(cf, f, 1);
	  goto finish;
	}
	
 	if (f->len < 4)
		goto drop;
	
	dest = node_lookup(f_getsymbol(f,2), 0);
	if (!dest)
		goto drop;

	DEBUG(9, printf("## send from %s to %s (%d)\n",
			name_lookup(f->nd), name_lookup(dest->name),
			f->mult););

	n = node_lookup(f->nd, 0);

	if (!seg_lookup(n,chan) || !seg_lookup(dest,chan)) goto drop;

	// J. Zivanov
	switch (dest->location)
	{
		case Local:
		
			node_lookup(f->nd, 0)->symbalance -= f->len + 3;
			node_lookup(dest->name, 0)->symbalance += f->len - 3;
		
			if (f->mult == 1) {
				f->offs += 3;
				f->len -= 3;
				f2 = f;
				f2->nd = dest->name;
			} else {
				f->mult--;
				f2 = f_new(f->buf, f->offs + 3, f->len - 3, 1, dest->name);
				t_append(f);
		//		addhint(node_lookup(f->nd, 0), SYM_SEND);
			}
			DEBUG(1, f_print(stdout, f2, 1););
			rx_fraglet(f2);
			
			break;
		
		case Remote:
		
			node_lookup(f->nd, 0)->symbalance -= f->len + 3;
		
			if (f->mult == 1) {
				f->offs += 3;
				f->len -= 3;
				f2 = f;
				f2->nd = dest->name;
			} else {
				f->mult--;
				f2 = f_new(f->buf, f->offs + 3, f->len - 3, 1, dest->name);
				t_append(f);
		//		addhint(node_lookup(f->nd, 0), SYM_SEND);
			}
			DEBUG(1, f_print(stdout, f2, 0); printf(" (unreliably)\n"););
			if (!packet_loss_probability || rand() % packet_loss_probability)
				send_fraglet_through_socket(f2, n, dest);
			else {DEBUG(1, printf("Loss emulator: OUTBOUND PACKET LOST\n"););}
			
			break;
	}
	
	return f;
drop:
	DEBUG(1, printf("[]\n"););
finish:
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_inject(struct frag *f)
{
	symbol child, parent;
	struct node *c, *p;
	int i;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;

	parent = f->nd;
	child = f_getsymbol(f,1);
	c = node_lookup(child, 0);
	p = node_lookup(parent, 0);
	// DEBUG(9, printf("## inject p=%d, c=%d\n", p, c););

	if (!p || !c) goto drop;

	for (i=0; i< p->nchildren; i++) {
	  if (p->child[i] == child) break;
	}

	if (i >= p->nchildren) {
	  DEBUG(9, printf("## inject from %s to %s failed: no such child\n",
			name_lookup(parent), name_lookup(child)););
	  goto drop;
	}

	DEBUG(9, printf("## inject from %s to %s (%d)\n",
			name_lookup(parent), name_lookup(child),
			f->mult););

	f = f_shrink(f,2);
	if (f) {
		f->nd = child;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;
drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_expel(struct frag *f)
{
	symbol child, parent;
	struct node *c, *p;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 2)
		goto drop;
	
	child = f->nd;
	c = node_lookup(child, 0);
	parent = c->parent;
	p = node_lookup(parent, 0);
	if (!p || !c) goto drop;

	DEBUG(9, printf("## expel from %s to %s (%d)\n",
			name_lookup(child), name_lookup(parent),
			f->mult););

	// (lidia):
	// TO DO: what to do with expel from root to root?
	// for the moment nothing happens
	f = f_shrink(f,1);
	if (f) {
		f->nd = parent;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;
drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_out(struct frag *f) 
{
	return f;
}

struct frag* f_lt(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_lt %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 5)
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 3]))
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 4]))
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		if (numberval(f->buf->data[f->offs + 3]) <
		    numberval(f->buf->data[f->offs + 4]))
			f->buf->data[f->offs + 2] = f->buf->data[f->offs + 1];
		f->offs += 2;
		f->len -= 2;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	if (numberval(f->buf->data[f->offs + 3]) <
	    numberval(f->buf->data[f->offs + 4]))
		buf->data[0] = f->buf->data[f->offs + 1];
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

// [eq yes no n m tail] --> if n==m: [yes n m tail], else: [no n m tail]
struct frag* f_eq(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_eq %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 5)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		if (f->buf->data[f->offs + 3] == f->buf->data[f->offs + 4])
			f->buf->data[f->offs + 2] = f->buf->data[f->offs + 1];
		f->offs += 2;
		f->len -= 2;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	if (f->buf->data[f->offs + 3] == f->buf->data[f->offs + 4])
		buf->data[0] = f->buf->data[f->offs + 1];
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_wait(struct frag *f)
{
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> ...\n"););
	      
	f = f_shrink(f, 0);
	if (f)
	  w_append(f, WAIT_DURATION);
	return f;
}


struct frag* f_pick(struct t_queue *q, int offs)
{
	struct frag *f, *prev;

/*
	printf("f_pick %d\n", offs);
	printf("\n* - - - - - - - - - - -\n");
	f_export(stdout);
	printf("\n* - - - - - - - - - - -\n");
*/


	// pick the fraglet at offset 'offs', taking its
	// multiplicity into account
	prev = 0;
	for (f = q->first; f; f = f->next) {
//		printf("f_pick mult=%d\n", f->mult);
		if (offs < f->mult)
			break;
		offs -= f->mult;
		prev = f;
	}
	
	if (!f) {
		fprintf(stderr, "# f_pick: no such fraglet\n");
		return 0;
	}

//	printf("f_pick found %d, next=%p\n", offs, f->next);

	// remove the fraglet from the queue
	if (!prev) {
		q->first = f->next;
	} else {
		prev->next = f->next;
	}
	if (q->last == f)
		q->last = prev;
	f->next = 0;

	return f;
}

int f_execFromQueue(struct node *n, struct t_queue *q, int offs)
{
	struct fragbuf *buf;
	struct frag *f, *prev, *f1, *f2, *f3;
	int ops = 0;
	int s;

	DEBUG(9, printf("## f_execFromQueue %d\n", offs););

	f = f_pick(q, offs);
	if (!f) {
		DEBUG(9, printf("## f_execFromQueue: no fraglet picked\n"););
		return 0;
	}

	s = f_peek(f);
	if (istrans(s)) {

		switch (s) {
		case SYM_NOP:	 f_nop(f);       break;
		case SYM_COPY:   f_copy(f);      break;
		case SYM_FORK:   f_fork(f);      break;
		case SYM_DELAY:  f_delay(f);	break;
		case SYM_DUP:    f_dup(f);       break;
		case SYM_EXCH:   f_exch(f);      break;
		case SYM_EMPTY:  f_empty(f);     break;
		case SYM_LT:	 f_lt(f);	break;
		case SYM_EQ:	 f_eq(f);	break;
		case SYM_LENGTH: f_length(f);	break;

		case SYM_SUB:
		case SYM_SUM:
		case SYM_MULT:
		case SYM_DIV:
		case SYM_MOD:
		case SYM_POW:    f_op2(f);	break;
		case SYM_ABS:    f_abs(f);	break;

		case SYM_NEWNAME:    f_newname(f);   break;
		case SYM_NEWNODE:    f_newnode(f);   break;
		case SYM_NUL:    f_nul(f);	break;
		case SYM_POP:	 f_pop(f);      break;
		case SYM_POP2:	 f_pop2(f);     break;
		case SYM_PRINTSYM:   f_printsym(f);     break;
		case SYM_SEND:   f_send(f);     break;
		case SYM_BCAST:  f_bcast(f);    break;
		case SYM_INJECT: f_inject(f);   break;
		case SYM_EXPEL:  f_expel(f);    break;
		case SYM_SPLIT:  f_split(f);    break;
		case SYM_SPLITAT:  f_splitat(f);     break;
		case SYM_WAIT:   f_wait(f);	break;
		case SYM_TRIGGER: appl_trigger(f); break;
		case SYM_OUT: f_out(f); break;
		default:
			fprintf(stderr, "'%s' not implemented\n",
			       name_lookup(s));
			if (--(f->mult) == 0)
				f_free(f);
			break;
		}
		return ops+1;
	}

	// printf("reaction ...\n");

	if (ismatch(s)) {
		struct o_rec *or;
		int i;
		symbol tag;
		f1 = f;

		// (lidia): match a number
		tag = f_getsymbol(f,1);
		if (isnumber(tag)) {
		  or = n->otab + SYM_NUMBER;
		  // (lidia): stopped here!
		} else {
		  or = n->otab + tag;
		}
		i = random() % or->active_cnt;

		f2 = f_pick((struct t_queue*) or, i);
		if (!f2) {
			rx_fraglet(f1);
			return 0;
		}
	} else {
		struct m_rec *mr;
		int i;

		f2 = f;
		mr = n->mtab + s;
		i = random() % mr->active_cnt;

		f1 = f_pick((struct t_queue*) mr, i);
		if (!f1) {
			rx_fraglet(f2);
			return 0;
		}
	}
	
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f1, 0); printf(", ");
	      f_print(stdout, f2, 0); printf(" --> "););
	      
	// the match generates a fraglet of length > 0
	if ((f1->len + f2->len - 3) > 0) {
		// allocate space for new fraglet and concatenate these
		// fraglets
		buf = malloc(sizeof(*buf) + sizeof(symbol) * (f1->len + f2->len - 3));
		buf->refcnt = 0;
		buf->len = f1->len + f2->len - 3;
		memcpy(buf->data, f1->buf->data + f1->offs + 2, sizeof(symbol)*(f1->len - 2));
		memcpy(buf->data + f1->len - 2, f2->buf->data + f2->offs + 1, sizeof(symbol)*(f2->len - 1));
		f3 = f_new(buf, 0, buf->len, 1, f1->nd);
		
		// print appropriate debug messages
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, f_print(stdout, f3, 1););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", ");
			      f_print(stdout, f3, 1););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
		node_lookup(f3->nd, 0)->symbalance -= f3->len;
		rx_fraglet(f3);
	} else {
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, fprintf(stdout, "[]\n"););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", []\n"););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
	}

	if (f1->buf->data[f1->offs] == SYM_MATCH) {
		node_lookup(f1->nd, 0)->symbalance += f1->len;
		f1->mult--;
	} // else SYM_MATCHP

	if (f1->mult <= 0)
		f_free(f1);
	else
		rx_fraglet(f1);

	node_lookup(f2->nd, 0)->symbalance += f2->len;
	f2->mult--;
	if (f2->mult <= 0)
		f_free(f2);
	else
		rx_fraglet(f2);

	return ops+1;
}




void f_cleanup()
{
	struct node *n;

	DEBUG(9, printf("## cleanup\n"););
	for (n = nodelist; n; n = n->next) {
		int i;
		struct m_rec *mtab;
		struct o_rec *otab;

		for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
			f_freeLinkedList((struct t_queue *) mtab);
		for (i = 0, otab = n->otab; i < MAXSYMBOLS; i++, otab++)
			f_freeLinkedList((struct t_queue *) otab);
		f_freeLinkedList(&(n->tq));

		// (lidia): should free waiting list too!
		f_freeLinkedList(&(n->wait));

		n->opcnt = 0;
		n->r_active = n->t_active = 0;
		//		n->hintcnt = 0;
	}

}


void collect_hints(struct node *n)
{
	int i;
	struct m_rec *mtab;
	struct o_rec *otab;
	struct frag *f;

	//DEBUG(9, printf("## collecthints %s\n", name_lookup(n->name)););

	/*
	if (n->tq.first) {
		if (addhint(n, f_peek(n->tq.first)))
			return;
	}
	for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
		if (mtab->first && n->otab[i].first) {
			if (addhint(n, i))
				return;
		}
	*/
	n->r_active = n->t_active = 0;
	for (i = 0, mtab = n->mtab, otab = n->otab;
				     i < MAXSYMBOLS; i++, mtab++, otab++) {
		mtab->active_cnt = 0;
		for (f = mtab->first; f; f = f->next) {
			if (otab->first)
				mtab->active_cnt += f->mult;
		}

		otab->active_cnt = 0;
		for (f = otab->first; f; f = f->next) {
			if (mtab->first)
				otab->active_cnt += f->mult;
		}
		mtab->r_active_cnt = mtab->active_cnt;
		if (mtab->active_cnt > otab->active_cnt)
			mtab->r_active_cnt = otab->active_cnt;

		n->r_active += mtab->r_active_cnt;
//		n->o_active += mtab->r_active_cnt;
	}
	for (f = n->tq.first; f; f = f->next)
		n->t_active += f->mult;
	n->total_active = n->r_active + n->t_active;
}


void run(int lim)
{
	int i;
	struct node *n;
	symbol s;
	int clock = 0;
	int anyactive;

	DEBUG(8, printf("# number size = %d\n", sizeof(number)););

	// J. Zivanov: if lim <=0 then infinite loop
	// (for continuous packet processing)

	while (lim <= 0 || lim > clock) {
		// if (lim > 0) {
		//  DEBUG(9, printf("## run loop %d/%d\n", clock, lim););
		//}
		
		anyactive = 0;
		for (n = nodelist; n; n = n->next) {
		
			if (n == nodelist) listen_on_local_sockets();

			// no processing of remote nodes:
			if (n->location == Remote) continue;

		  	// pending: modify "wait" to take into accout
		  	// "real" time instead of virtual time

			if (n->wait.first && n->opcnt >= n->wait.first->expires) {
				struct frag *f = n->wait.first;
				n->wait.first = f->next;
				if (n->wait.last == f)
					n->wait.last = 0;
				f->next = 0;
				if (f->buf->data[f->offs] == SYM_WAIT)
				  f = f_shrink(f,1);
				else if (f->buf->data[f->offs] == SYM_DELAY)
				  f = f_shrink(f,2);
				else {
				  fprintf(stderr, "** bad: unknown head in wait queue\n");
				  exit(-1);
				}
				DEBUG(1,
				      printf("%05d  ",
					     node_lookup(f->nd, 0)->opcnt);
				      printf("... --> ");
				      f_print(stdout, f, 1););

				if (f)
					rx_fraglet(f);
				n->possibleHint |= 1;
			}

			collect_hints(n);
			DEBUG(9, printf("## run loop %d/%d: node=%s tactive=%d twait=%lu\n", clock, lim, name_lookup(n->name), n->total_active, n->wait.first););

			if (n->total_active) {
				anyactive = 1;
				i = random() % n->total_active;
				DEBUG(8, printf("# i=%d from %d, r_act=%d\n", i, n->total_active, n->r_active););
				if (i < n->r_active) {
					struct m_rec *mr;
					mr = n->mtab;
					while (i >= mr->r_active_cnt) {
						i -= mr->r_active_cnt;
						mr++;
						DEBUG(8, printf("while i=%d act=%d symb=%s\n",i, mr->r_active_cnt, name_lookup(mr - n->mtab)););
					}
					DEBUG(8, printf("picking %d mtab[%s]\n", i, name_lookup(mr - n->mtab)););
					i = random() % mr->active_cnt;
					DEBUG(8, printf("within mtab[%s], picking %d out of %d \n", name_lookup(mr - n->mtab), i, mr->active_cnt););
					f_execFromQueue(n, (struct t_queue *)mr, i);
				} else {
						i -= n->r_active;
						f_execFromQueue(n, &n->tq, i);
				}
			} else
				if(n->wait.first)
					anyactive = 1;
				else 
					anyactive |= 0;

			n->opcnt++;
		}
		clock++;

	  // not good place: select ( timeout ) // udp socket ??? or > 1 socket
	  
	  // no more fraglets are active, so don't wait until maximum number of steps is 
	  // finished; this may be a waste of time!
	  if(!anyactive)
	  	break;

	}
}

int main(int argc, char *argv[])
{
	int lim = 100;
	int soffs = 0;
	repeatable = 0;

	init_sockets();

	for (argc--, argv++; argc > 0; argc--, argv++) {
		if (!strcmp(argv[0], "-d") && argc > 1) {
			debug = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-e") && argc > 1) {
			dump = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-lim") && argc > 1) {
			lim = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-soffs") && argc > 1) {
			soffs = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-lose") && argc > 1) {
			packet_loss_probability = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-r")) {
			repeatable = 1;
			continue;
		}
		if (!strcmp(argv[0], "-q")) {
			quiet = 1 - quiet;
			continue;
		}
		if (!strcmp(argv[0], "-h") ||
		    !strcmp(argv[0], "-help") ||
                    !strcmp(argv[0], "--help")) goto usage;
		
		fprintf(stderr, "unknown option %s\n", argv[0]);
	usage:
		fprintf(stderr, "usage: frag [options]\n");
		fprintf(stderr, "  -d lev     // debug\n");
		fprintf(stderr, "  -e N       // export: 0=none, 1=in, 2=out, 3=both\n");
		fprintf(stderr, "  -lim N     // limit exec count\n");
		fprintf(stderr, "  -soffs N   // score index offset\n");
		fprintf(stderr, "  -lose N    // udp loss probability 1/N\n");
		fprintf(stderr, "  -r         // repeatable (default no)\n");
		fprintf(stderr, "  -q         // quiet (no header output)\n");
		exit(-1);
	}

	// (lidia) be careful to add symbols in the same order as frag.h
	// enum SYM_*
	name_add("nop");
	name_add("copy");
	name_add("fork");
	name_add("delay");
	name_add("dup");
	name_add("empty");
	name_add("exch");
	name_add("lt");
	name_add("eq");
	name_add("length");
	name_add("sum");
	name_add("sub");
	name_add("mult");
	name_add("div");
	name_add("mod");
	name_add("abs");
	name_add("pow");
	name_add("newname");
	name_add("newnode");
	name_add("nul");
	name_add("pop");
	name_add("pop2");
	name_add("printsym");
	name_add("send");
	name_add("broadcast");
	name_add("inject");
	name_add("expel");
	name_add("split");
	name_add("splitat");
	name_add("wait");
	name_add("_");
	name_add("match");
	name_add("matchp");
	name_add("*");
	name_add("stdin");
	name_add("stdout");
	name_add("stderr");
	name_add("out");
	name_add("__"); // (lidia): (tmp) placeholder for all numbers
	// SYM_EOL is end-of-list marker

	appl_init();
	while (!f_import(stdin)) {
		appl_scoreReset();
		// Lidia: reset seed for random(), if repeatability needed
		if (repeatable) srandom(1);

		if (dump || debug) {
			time_t t;
			struct passwd *u;
			struct utsname un;

			time(&t);
			u = getpwuid(getuid());
			uname(&un);

			if (!quiet) {
				printf("# " VERSION);
				printf("# Executed by %s on %s at %s",
				       u->pw_name, un.nodename, ctime(&t));
			}
		}

		if (dump == 1 || dump == 3) {
			if (!quiet)
				printf("\n# - - - - INITIAL MULTISET: - - - - - - -\n");
			f_export(stdout);
		}
		if (debug && !quiet)
			printf("\n# - - - - EXECUTION: - - - -- - - - - - -\n");

		run(lim);

		if (dump == 2 || dump == 3) {
			if (!quiet)
				printf("\n# - - - - END MULTISET: - - - - - - - - -\n");
			f_export(stdout);
		}

		if (lim > 0) {
			if ((dump || debug) && !quiet) {
				printf("\n# - - - - - - - - - - -\n");
			//	printf("# score: %d\n", finalScore());
				printf("s %d %d\n", appl_finalScore(), ++soffs);
			}
		}

		f_cleanup();
	}
	return 0;
}

#include "sockin.c"

// eof
