/* fraglets0.30/src/frag.h

  header file for fraglet interpreter core, frag.c

  Copyright (C) 2003-2007 Christian Tschudin <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  History:

  20070924  cft: added printsym operator
  20070924  lidia: added abs operator, renamed 'diff' to 'sub' (subtract)
  20070921  lidia: added stdin, stdout and stderr symbols
  20070919  lidia: added broadcast instruction and mod (%) operator
  20070805  lidia: added child/parent structure for nodes, plus
            "inject" and "expel" fraglets to send a fraglet to a child
	    and to the parent, respectively.
  20070720  thomas: added "fork" and "eq" instructions
  20070320  lidia: added SPLITAT transformation
  20070314  daniel.schreckling@informatik.uni-hamburg.de: added support for 
            negative numbers
  20070312  lidia: added SYM_POW (x^y) (requires math.h for pow())
  20070303  lidia: added SYM_MULT and SYM_DIV ("mult" and "div" operations)
            increased symbol size to long
  20070110  lidia: added node creation ("newnode")
  20061213  christian.tschudin@unibas.ch: added number support
  20050526  Jasenko Zivanov (Unibas): added socket support
  20050204  lidia.yamamoto@unibas.ch: added SYM_DEL for 'nul' op as in
            ains2003 (fraglets v0.17)
  20040704  updated by christian.tschudin@unibas.ch (fraglets v0.16)
  2003????  created by christian.tschudin@unibas.ch

*/

#ifndef FRAG_EE_H
#define FRAG_EE_H


#include <ctype.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>
#include <math.h>

#define VERSION "frag v0.29 (Fraglet interpreter)\n# (c) 2003-2007 <christian.tschudin@unibas.ch>\n"

#define SYM_NUMBEROFFS (1<<13)
#define MAXSYMBOLS   (SYM_NUMBEROFFS-1)
#define MAXNODES     (1<<5) // max number of inner nodes per node
#define MAXSEGMENTS  (1<<5)
#define MAXHINTS     50

enum { SYM_NOP = 1,
       SYM_COPY,
       SYM_FORK,
       SYM_DELAY,
       SYM_DUP,
       SYM_EMPTY,
       SYM_EXCH,
       SYM_LT,
       SYM_EQ,
       SYM_LENGTH,
       SYM_SUM,
       SYM_SUB,
       SYM_MULT,
       SYM_DIV,
       SYM_MOD,
       SYM_ABS,
       SYM_POW,
       SYM_NEWNAME,
       SYM_NEWNODE,
       SYM_NUL,
       SYM_POP,
       SYM_POP2,
       SYM_PRINTSYM,
       SYM_SEND,
       SYM_BCAST,
       SYM_INJECT,
       SYM_EXPEL,
       SYM_SPLIT,
       SYM_SPLITAT,
       SYM_WAIT,
       SYM_TRIGGER,
       // ---------
       SYM_MATCH,
       SYM_MATCHP,
       SYM_STAR,
       SYM_STDIN,
       SYM_STDOUT,
       SYM_STDERR,
       // there is almost no way to not add an 
       // "out" symbol without being too inefficient
       SYM_OUT,
       SYM_NUMBER,
       SYM_EOL
       // careful!! SYM_EOL actually marks the first user-defined symbol!!
};
#define SYM_LAST_TRANS SYM_TRIGGER
#define SYM_LAST_OP SYM_EOL

#define WAIT_DURATION 10
//#define WAIT_DURATION 40

// (lidia:) note: numbers and symbols must have the same size since
// they're put in the same buffer that represents a fraglet (sequence
// of symbols), see struct fragbuf below
//

#if 0
#define T_SYM_SHORT
// (daniel:) note: if symbol is short then with the current 
// implementationwe can only use positive numbers as there is no room
// for negatives
typedef unsigned short symbol;
typedef unsigned short number;
#define NEGNUMBEROFFS 0
#else
#define T_SYM_LONG
typedef int number;
typedef unsigned long symbol;
// (daniel:) note: with symbol size long you can at most use ints as 
// numbers otherwise you will overflow the symbol size (this assumes that
// sizeof(int) <= sizeof(long) which is not necessarily true!
#define NEGNUMBEROFFS ((1 << (sizeof(int)*sizeof(char)*8 - 1)) + 1)
#endif

struct fragbuf {
	unsigned short refcnt;
	unsigned short len;         // in symbols
	symbol data[0];
};

struct frag {
	struct fragbuf *buf;
	struct frag *next;
	symbol nd;   // node where fraglet resides
	int mult;  // multiplicity counter
	short offs, len; // in symbols
	char *cp;
	int expires;
};

struct m_rec {
	struct frag *first;
	struct frag *last;
	int active_cnt;
	int r_active_cnt;
};


struct o_rec {
	struct frag *first;
	struct frag *last;
	int active_cnt;
	int dummy;
};

struct t_queue {
	struct frag *first;
	struct frag *last;
	int active_cnt;
	int dummy;
};

typedef enum NodeLocation
{ Local, Remote } NodeLocation;

struct UDPNode;

struct node {
	struct node *next;
	symbol name, parent, child[MAXNODES];
	symbol segments[MAXSEGMENTS];
	int nchildren;

   // newtypeTBD attach[MAXSEGMENTS];  "prot ip port"

	// list of fraglets matching a specific symbol
	struct m_rec mtab[MAXSYMBOLS];
	
	struct o_rec otab[MAXSYMBOLS];
	struct t_queue tq;
	struct t_queue wait;
	int r_active, t_active;
	int total_active;
	int opcnt;
	int symbalance;
	struct UDPNode* udp_wrapper;
	NodeLocation location;
	/*int socketFD;
	char ipAddress[16];*/
//	int hintcnt;
//	symbol hints[MAXHINTS];
	char possibleHint;
};

struct name {
	struct name *next;
	char *cp;
	symbol sym;
};


extern void appl_init();
extern void appl_fini();
extern  int appl_clockEvent();
extern void appl_trigger(struct frag *f);
extern void appl_scoreReset();
extern  int appl_finalScore();

#endif //FRAG_EE_H
